@@ -100,6 +100,7 @@ from pyvx.backend import *
 import cffi
 import os
 
+
 class ElementwiseNode(Node):
     small_ints = ('uint8_t', 'int8_t', 'uint16_t', 'int16_t')
 
@@ -112,7 +113,8 @@ class ElementwiseNode(Node):
             img.ensure_similar(inputs[0])
             if self.convert_policy == CONVERT_POLICY_SATURATE:
                 if img.image_format.ctype not in self.small_ints:
-                    raise ERROR_INVALID_FORMAT("Saturated arithmetic only supported for 8- and 16- bit integers.")
+                    raise InvalidFormatError(
+                        "Saturated arithmetic only supported for 8- and 16- bit integers.")
 
     def tmptype(self, ctype):
         if ctype in self.small_ints:
@@ -122,7 +124,7 @@ class ElementwiseNode(Node):
     def compile(self, code, noloop=False):
         iin = self.input_images.items() + self.inout_images.items()
         iout = self.output_images.items() + self.inout_images.items()
-        magic = {'__tmp_image_%s' % name : img 
+        magic = {'__tmp_image_%s' % name: img
                  for name, img in iin + iout}
         setup = ''.join("%s %s;" % (self.tmptype(img.image_format.ctype), name)
                         for name, img in iin + iout)
@@ -133,29 +135,35 @@ class ElementwiseNode(Node):
         if noloop:
             head = ""
         else:
-            head = "for (long __i = 0; __i < __tmp_image_%s.values; __i++) " % iin[0][0]
+            head = "for (long __i = 0; __i < __tmp_image_%s.values; __i++) " % iin[
+                0][0]
         body = inp + self.body + outp
         block = head + "{" + body + "}"
         code.add_block(self, setup + block, **magic)
 
+
 class MergedElementwiseNode(MergedNode):
+
     def compile(self, code):
         img = self.original_nodes[0].input_images.values()[0]
         code.add_code("\n// MergedElementwiseNode\n")
         code.indent_level += 4
         code.add_code("for (long __i = 0; __i < %s; __i++) {\n" %
-                       img.getattr(self, "values"))
+                      img.getattr(self, "values"))
         for n in self.original_nodes:
             n.compile(code, True)
         code.indent_level -= 4
         code.add_code("}\n")
 
+
 class BinaryOperationNode(ElementwiseNode):
     signature = "in in1, in op, in in2, in convert_policy, out out"
+
     @property
     def body(self):
         return "out = in1 %s in2;" % self.op
 
+
 class MultiplyNode(ElementwiseNode):
     signature = "in in1, in in2, in scale, in convert_policy, in round_policy, out out"
     scale = 1
@@ -170,6 +178,7 @@ class MultiplyNode(ElementwiseNode):
         else:
             raise NotImplementedError
 
+
 class DivideNode(ElementwiseNode):
     signature = "in in1, in in2, in scale, in convert_policy, in round_policy, out out"
     scale = 1
@@ -184,6 +193,7 @@ class DivideNode(ElementwiseNode):
         else:
             raise NotImplementedError
 
+
 class TrueDivideNode(ElementwiseNode):
     signature = "in in1, in in2, out out"
     body = "out = ((double) in1) / ((double) in2);"
@@ -192,10 +202,12 @@ class TrueDivideNode(ElementwiseNode):
         self.out.suggest_color(DF_IMAGE_F64)
         ElementwiseNode.verify(self)
 
+
 class PowerNode(ElementwiseNode):
     signature = "in in1, in in2, in convert_policy, out out"
     body = "out = pow(in1, in2);"
 
+
 class CompareNode(BinaryOperationNode):
     signature = "in in1, in op, in in2, out out"
 
@@ -203,24 +215,86 @@ class CompareNode(BinaryOperationNode):
         self.out.suggest_color(DF_IMAGE_U8)
         ElementwiseNode.verify(self)
 
+
+class ColorConvertNode(Node):
+    signature = 'in input, out output'
+    convert_policy = CONVERT_POLICY_SATURATE
+
+    def verify(self):
+        for im in (self.input, self.output):
+            if im.color_space != COLOR_SPACE_DEFAULT:
+                raise NotImplementedError
+            if im.channel_range != CHANNEL_RANGE_FULL:
+                raise NotImplementedError
+        self.output.ensure_shape(self.input)
+
+    def compile(self, code):
+        in_channels = self.input.image_format.channels
+        out_channels = self.output.image_format.channels
+        if CHANNEL_R in in_channels and CHANNEL_Y in out_channels:
+            code.add_block(self, """
+                double kr = 0.2126, kb = 0.0722, kg = 1.0 - kr - kb;
+                for (long i = 0; i < out.pixels; i++) {
+                    double r = ((double)input.channel_r[i]) / 256.0;
+                    double g = ((double)input.channel_g[i]) / 256.0;
+                    double b = ((double)input.channel_b[i]) / 256.0;
+                    double y = kr*r + kg*g + kb*b;
+                    double u = b/2.0 - (kr*r + kg*g) / (2.0 - 2.0*kb);
+                    double v = r/2.0 - (kb*b + kg*g) / (2.0 - 2.0*kr);
+                    out.channel_y[i] = y * 256.0;
+                    out.channel_u[i] = floor(u * 256.0 + 128.0);
+                    out.channel_v[i] = floor(u * 256.0 + 128.0);
+                }
+            """, out=self.output, input=self.input)
+        else:
+            raise NotImplementedError
+
+
 class ChannelExtractNode(Node):
     signature = "in input, in channel, out output"
 
     def verify(self):
         if self.channel not in self.input.image_format.channels:
             raise InvalidFormatError(
-                'Cant extract channel %d from %s image.' % (
-                    channel_number[self.channel], self.input.image_format.__name__))
-        self.output.ensure_color(DF_IMAGE_U8)        
+                'Cant extract CHANNEL_%s from %s image.' % (
+                    channel_char[self.channel].upper(), self.input.image_format.__name__))
+        self.output.ensure_color(DF_IMAGE_U8)
         self.output.ensure_shape(self.input)
 
     def compile(self, code):
         code.add_block(self, """
             for (long i = 0; i < out.pixels; i++) {
-                out[i] = input.channel_%d[i];
+                out[i] = input.channel_%s[i];
             }
-            """ % channel_number[self.channel], 
-            input=self.input, out=self.output)
+            """ % channel_char[self.channel],
+                       input=self.input, out=self.output)
+
+
+class ChannelCombineNode(Node):
+    signature = 'in plane0, in plane1, in plane2, in plane3, out output'
+    plane2 = None
+    plane3 = None
+
+    def verify(self):
+        self.output.suggest_color(DF_IMAGE_RGB)
+        if self.output.color != DF_IMAGE_RGB:
+            raise NotImplementedError
+        if self.plane2 is None or self.plane3 is not None:
+            raise InvalidNodeError('RGB image requires 3 planes.')
+        self.output.ensure_shape(self.plane0)
+        self.plane1.ensure_shape(self.plane0)
+        self.plane2.ensure_shape(self.plane0)
+
+    def compile(self, code):
+        code.add_block(self, """
+            for (long i = 0; i < out.pixels; i++) {
+                out.channel_r[i] = red[i];
+                out.channel_g[i] = green[i];
+                out.channel_b[i] = blue[i];
+            }
+            """, red=self.plane0, green=self.plane1, blue=self.plane2, out=self.output)
+
+
 
 class Gaussian3x3Node(Node):
     signature = "in input, out output"
@@ -240,6 +314,7 @@ class Gaussian3x3Node(Node):
             }
             """, img=self.input, res=self.output)
 
+
 class Sobel3x3Node(Node):
     signature = 'in input, out output_x, out output_y'
 
@@ -267,6 +342,7 @@ class Sobel3x3Node(Node):
             }
             """, img=self.input, dx=self.output_x, dy=self.output_y)
 
+
 class MagnitudeNode(ElementwiseNode):
     signature = 'in grad_x, in grad_y, out mag'
     convert_policy = CONVERT_POLICY_SATURATE
@@ -285,6 +361,7 @@ class MagnitudeNode(ElementwiseNode):
 
     body = "mag = sqrt( grad_x * grad_x + grad_y * grad_y );"
 
+
 class PhaseNode(ElementwiseNode):
     signature = 'in grad_x, in grad_y, out orientation'
 
@@ -297,6 +374,7 @@ class PhaseNode(ElementwiseNode):
 
     body = "orientation = (atan2(grad_y, grad_x) + M_PI) * (255.0 / 2.0 / M_PI);"
 
+
 class AccumulateImageNode(Node):
     signature = 'in input, inout accum'
 
@@ -305,6 +383,7 @@ class AccumulateImageNode(Node):
 
 mydir = os.path.dirname(os.path.abspath(__file__))
 
+
 class PlayNode(Node):
     signature = 'in path, out output'
     player = None
@@ -328,10 +407,9 @@ class PlayNode(Node):
     except (cffi.VerificationError, IOError) as e:
         lib = None
 
-
     def verify(self):
         if self.lib is None:
-            raise ERROR_INVALID_NODE('''
+            raise InvalidValueError('''
                                      
                 PlayNode failed to compile. See error message from the compiler above,
                 and make sure you have vlc installed. On Debian:
@@ -343,25 +421,27 @@ class PlayNode(Node):
 
                     pip install --upgrade --force-reinstall pyvx
 
-                ''')        
+                ''')
         if not self.player:
             self.player = self.lib.vlcplay_create(self.path)
         if not self.player:
-            raise ERROR_INVALID_VALUE("Unable to decode '%s' using vlc." % self.path)
+            raise InvalidValueError(
+                "Unable to decode '%s' using vlc." % self.path)
         self.output.ensure_shape(self.player.width, self.player.height)
         self.output.ensure_color(DF_IMAGE_RGB)
         self.output.force()
 
     def compile(self, code):
         adr = int(self.ffi.cast('long', self.player))
-        code.add_block(self, "if (vlcplay_next((void *)0x%x, img.data)) return VX_ERROR_GRAPH_ABANDONED;" % adr, img=self.output);
+        code.add_block(
+            self, "if (vlcplay_next((void *)0x%x, img.data)) return VX_ERROR_GRAPH_ABANDONED;" % adr, img=self.output)
         code.extra_link_args.append(self.ffi.verifier.modulefilename)
         code.includes.add('#include "vlcplay.h"')
 
-
     def __del__(self):
         self.lib.vlcplay_release(self.player)
 
+
 class ShowNode(Node):
     signature = "in input, in name"
     viewer = None
@@ -382,7 +462,7 @@ class ShowNode(Node):
 
              """)
     try:
-        lib = ffi.verify(open(os.path.join(mydir, 'glview.c')).read(), 
+        lib = ffi.verify(open(os.path.join(mydir, 'glview.c')).read(),
                          extra_compile_args=['-O3'],
                          libraries=['glut', 'GL', 'GLU'])
     except (cffi.VerificationError, IOError) as e:
@@ -391,7 +471,7 @@ class ShowNode(Node):
 
     def verify(self):
         if self.lib is None:
-            raise ERROR_INVALID_NODE('''
+            raise InvalidValueError('''
 
                 ShowNode failed to compile. See error message from the compiler above,
                 and make sure you have glut GL and GLU installed. On Debian:
@@ -409,15 +489,16 @@ class ShowNode(Node):
             if self.viewer.width == self.input.width and self.viewer.height == self.input.height:
                 return
             self.lib.glview_release(self.viewer)
-        self.viewer = self.lib.glview_create(self.input.width, 
+        self.viewer = self.lib.glview_create(self.input.width,
                                              self.input.height,
-                                             self.lib.GL_RGB, 
-                                             self.lib.GL_UNSIGNED_BYTE, 
+                                             self.lib.GL_RGB,
+                                             self.lib.GL_UNSIGNED_BYTE,
                                              self.name)
 
     def compile(self, code):
         adr = int(self.ffi.cast('long', self.viewer))
-        code.add_block(self, "if (glview_next((void *)0x%x, img.data)) return VX_ERROR_GRAPH_ABANDONED;" % adr, img=self.input);
+        code.add_block(
+            self, "if (glview_next((void *)0x%x, img.data)) return VX_ERROR_GRAPH_ABANDONED;" % adr, img=self.input)
         code.extra_link_args.append(self.ffi.verifier.modulefilename)
         code.includes.add('#include "glview.h"')
 
